﻿module WebTools

open System
open System.Drawing 
open System.Windows.Forms
open System.Text
open System.Net
open System.IO
open System.Text.RegularExpressions
open System.Runtime.Serialization
open System.Collections
open System.Collections.Generic  
open System.Xml.Linq
open System.Xml.XPath 

open Resoures

type DelegateAddNodeToTreeView = delegate of System.Windows.Forms.TreeView*string*string -> System.Windows.Forms.TreeNode
type DelegateAddNodeToNode = delegate of System.Windows.Forms.TreeNode*string*string -> System.Windows.Forms.TreeNode

type Operation =
    | None
    | AddChildNodes
    | DownloadAll
    | BrowseAll
    | ReplaceAll

[<AllowNullLiteral>]
type StatusStripValues() as this =
    let mutable m_statusLabelValue: string = ModuleResoures.GetAppResourceString("StatusReady") 
    let mutable m_progressBarValue: int = 0
    let mutable m_curCount: int = 0
    let mutable m_maxCount: int = 0
    let mutable m_operationCompleted: bool = false
    let mutable m_operationTerminated: bool = false
    let mutable m_lstErrors: (string*string) list = []

    do this.Reset()

    member this.StatusLabelValue
        with get() = m_statusLabelValue
        and set(value : string) = m_statusLabelValue <- value
    member this.ProgressBarValue
        with get() = m_progressBarValue
        and set(value : int) = m_progressBarValue <- value
    member this.CurCount
        with get() = m_curCount
        and set(value : int) = m_curCount <- value
    member this.MaxCount
        with get() = m_maxCount
        and set(value : int) = m_maxCount <- value
    member this.OperationCompleted
        with get() = m_operationCompleted
        and set(value : bool) = m_operationCompleted <- value
    member this.OperationTerminated
        with get() = m_operationTerminated
        and set(value : bool) = m_operationTerminated <- value
    member this.ListErrors
        with get() = m_lstErrors
        and set(lstErrors : (string*string) list) = m_lstErrors <- lstErrors @ m_lstErrors
    member this.Reset() =
        this.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusReady")
        this.ProgressBarValue <- 0
        this.CurCount <- 0
        this.MaxCount <- 0
        this.OperationCompleted <- false
        this.OperationTerminated <- false
        this.ListErrors <- []

(*
let SerializeObj<'t>(clsObj: obj, file: string) = 
    let ser: DataContractSerializer = new DataContractSerializer(typeof<'t>)
    use writer: FileStream = new FileStream(file, FileMode.Create)
    ser.WriteObject(writer, clsObj)

let DeserializeObj<'t>(file: string) = 
    let ser: DataContractSerializer = new DataContractSerializer(typeof<'t>)
    use reader: FileStream = new FileStream(file, FileMode.Open)
    ser.ReadObject(reader) :?> 't
*)

let xname thing = XName.Get(thing.ToString())
 
type xmlTag =
    | Link of string*string
    | GrpLink of string*((string*string) list) 
    | GrpLinkEx of string*(xmlTag list) 
    | File of string
    member public this.ToXml() =
        match this with
        | Link(href: string, title: string) ->
            new XElement(xname "link",
                new XElement(xname "href", href),
                new XElement(xname "title", title))
        | GrpLink(grpType: string, lstLink: (string*string) list) ->
            new XElement(xname grpType, 
                seq {for p in lstLink ->
                        match p with
                            | (h, t) ->
                                    let (h, t) = p
                                    let l = Link(h, t)
                                    l.ToXml()
                     })
        | GrpLinkEx(grpType: string, lstLink: xmlTag list) ->
            new XElement(xname grpType, 
                seq {for p in lstLink ->
                        match p with
                        | Link(h, t) ->
                            let l = Link(h, t)
                            l.ToXml()
                        | _ -> raise <| new ArgumentException(p.GetType().ToString() + " is not type Link(h, t)")})
        | File(file: string) ->
            new XElement(xname "file", file)

type xmlAdditional =
    | AdditionalLink of string*string
    member public this.ToXml() =
        match this with
        | AdditionalLink(hrefCanonical: string, hrefNormalized: string) ->
            new XElement(xname "link",
                new XElement(xname "hrefCanonical", hrefCanonical),
                new XElement(xname "hrefNormalized", hrefNormalized))

let convToXml(lst: xmlTag list) =
    let res = lst
            |> List.map(fun o -> o.ToXml())
            |> Array.ofList
    new XElement(xname "root", res)

let GetLocalFile(folder: string, localFile: string, fileExt: string): string =   
    Path.Combine(folder, Path.GetFileNameWithoutExtension(localFile) + fileExt) 

let GetFilesInFolder(folder: string): string[] =
    folder
    |> Directory.GetFiles
    |> Array.map Path.GetFileName 

let InitHashSetFilesInFolder(folder: string): HashSet<string> =
    new HashSet<string>(GetFilesInFolder(folder))

let GetLocalFileForUrl(url: string): string =
    // let regExprAspExt = ".asp[x]{0,1}"
    let fn = Path.GetFileName(url)
    if (fn.Contains(".asp")) then
        Regex.Replace(fn, CommonDefs.RegExprAspExt, CommonDefs.HtmlExt)
    else
        let lpos = fn.LastIndexOf('.')
        if (lpos = (-1)) then
            fn + CommonDefs.HtmlExt
        else
            fn

let GetResFileName(folder: string, url: string): string =
    let rec GetResFileNameRec(url, lstRepl) =
        match lstRepl with
        | [] -> url
        | head::tail -> GetResFileNameRec(Regex.Replace(url, head, String.Empty), tail)
    let posAmp = url.LastIndexOf(CommonDefs.HtmlAmp)
    let fn = GetResFileNameRec((if(posAmp <> (-1)) then url.Substring(0, posAmp) else url), CommonDefs.NotFileNameParts)
    if ((fn.Length + folder.Length + CommonDefs.SubfolderFiles.Length) > CommonDefs.MaxFileNameLength) then 
        fn.Substring(0, CommonDefs.MaxFileNameLength - (folder.Length + CommonDefs.SubfolderFiles.Length + 5)) 
    else 
        fn

let GetWebRequest(uri: Uri): HttpWebRequest =
    let wp = WebRequest.GetSystemWebProxy();
    wp.Credentials <- CredentialCache.DefaultCredentials;
    let req = WebRequest.Create(uri) :?> HttpWebRequest
    req.Proxy <- wp
    req

let GetUrlContents(url: string, lang: string): string =
    let uri = new Uri(url)
    let req = GetWebRequest(uri)
    if (not <| String.IsNullOrEmpty(lang)) then
        let cookieLang = CommonDefs.CookieLangName + "=" + lang + "; "
                        + "path=" + uri.Host + "/" 
        req.Headers.Add(HttpRequestHeader.Cookie, cookieLang)
    use stream = req.GetResponse().GetResponseStream()
    use reader = new IO.StreamReader(stream)
    reader.ReadToEnd()

let RemoveBracesContentsFromUrl(url: string): string =
    // let regExprInBraces = "\((\s*\S*)?\)"
    let reInBraces = new Regex(CommonDefs.RegExprInBraces, RegexOptions.IgnoreCase)
    let matchInBraces = reInBraces.Match(url)
    if (matchInBraces.Success && matchInBraces.Groups.Count > 0) then
        Regex.Replace(url, "\(" + matchInBraces.Groups.Item(1).Captures.Item(0).Value + "\)", String.Empty)
    else
        url 

let CorrectUrl(purl: string, url: string): string =
    // let regExprHttp = "http[s]{0,1}://\s*\S*"
    if (not <| Regex.Match(url, CommonDefs.RegExprHttp).Success) then 
        let puri = new Uri(purl)
        if (url.[0..1] = "//") then  
            puri.Scheme + ":" + url
        else     
            puri.Scheme + "://" + puri.Host + url
    else url

let DownloadUrlToFile(purl: string, url: string, file: string) =
    let uri = CorrectUrl(purl, url)
    let req = GetWebRequest(new Uri(uri))
    req.KeepAlive <- false
    use stream = req.GetResponse().GetResponseStream()
    use fileStream = File.OpenWrite(file)
    stream.CopyTo(fileStream) 
    (*
    let mutable buff: byte[] = Array.zeroCreate(CommonDefs.BuffSize)
    let mutable lenRead = stream.Read(buff, 0, CommonDefs.BuffSize)
    while (lenRead > 0) do 
        fileStream.Write(buff, 0, lenRead)
        lenRead <- stream.Read(buff, 0, CommonDefs.BuffSize)
    *)

let GetFileContents (file: string): string =
    use sr = File.OpenText(file)
    sr.ReadToEnd()

let SaveAsFile(file: string, data: string) = 
    use sr = File.CreateText(file)
    sr.Write(data)

let SaveAsFileEx(file: string, data: byte[]) = 
    use sr = new StreamWriter(file)
    sr.Write(data)

let ApplyReplaceLstExprs(html: string, lstReplacement: (string*string) list): string =
    let mutable retVal = html
    for (rexpr, text) in lstReplacement do
        retVal <- Regex.Replace(retVal, rexpr, text)
    retVal

let GetCanonicalUrlForHtmlDoc(html: string): string =
    //let regExprLinkCanonical = "<\s*link\s+[\s\S]*?rel=\"canonical\"([\s\S]*?)\s+/\s*>"
    //let regExprAttrHref = "href\s*=\s*\"([^\"]*)\""
    let reLinkCanonical = new Regex(CommonDefs.RegExprLinkCanonical, RegexOptions.IgnoreCase)
    let matchLinkCanonical = reLinkCanonical.Match(html)
    if (matchLinkCanonical.Success && matchLinkCanonical.Groups.Count > 0) then
        //let hrefExpr = Regex.Replace(Regex.Replace(matchLinkCanonical.Groups.Item(1).Captures.Item(0).Value, "\(", "\\("), "\)", "\\)") 
        let hrefExpr = matchLinkCanonical.Groups.Item(1).Captures.Item(0).Value
        Regex.Match(hrefExpr, 
            CommonDefs.RegExprAttrHref).Groups.Item("1").Captures.Item(0).Value 
    else        
        String.Empty

let GetNormalizedUrlForHtmlDoc(html: string): string =
    //let regExprMetaNormalizedUrl = "<\s*meta\s+[\s\S]*?name=\"NormalizedUrl\"([\s\S]*?)\s+/\s*>"
    //let regExprAttrContent = "content\s*=\s*\"([^\"]*)\""
    let reMetaNormalizedUrl = new Regex(CommonDefs.RegExprMetaNormalizedUrl, RegexOptions.IgnoreCase)
    let matchMetaNormalizedUrl = reMetaNormalizedUrl.Match(html)
    if (matchMetaNormalizedUrl.Success && matchMetaNormalizedUrl.Groups.Count > 0) then
        let contentExpr = matchMetaNormalizedUrl.Groups.Item(1).Captures.Item(0).Value
        Regex.Match(contentExpr, 
            CommonDefs.RegExprAttrContent).Groups.Item("1").Captures.Item(0).Value 
    else        
        String.Empty
         
let GetUrlForHtmlDoc(html: string): string =
    let normalizedUrl = GetNormalizedUrlForHtmlDoc(html)
    if (String.IsNullOrEmpty(normalizedUrl)) then
        GetCanonicalUrlForHtmlDoc(html)
    else
        RemoveBracesContentsFromUrl(normalizedUrl)

let GetHtmlBeforeBody(html: string): string =
    //let regExrBody = "<\s*body(\s[^>]*)?>[\s\S]*?<\s*/body\s*>"
    let reBody = new Regex(CommonDefs.RegExprBody, RegexOptions.IgnoreCase)
    let matchBody = reBody.Match(html)
    if (matchBody.Success) then
        let partHtml = html.Substring(0, matchBody.Index)
        //let regExprHeadClose = "<\s*/\s*head\s*>"
        let reHeadClose = new Regex(CommonDefs.RegExprHeadClose, RegexOptions.IgnoreCase)
        let matchHeadClose = reHeadClose.Match(partHtml)
        if (matchHeadClose.Success) then
            partHtml.Substring(0, matchHeadClose.Index) 
                + CommonDefs.StrNewLine + CommonDefs.StyleExternalImage + CommonDefs.StrNewLine 
                + matchHeadClose.Value 
        else
            partHtml
    else 
         raise <| new NullReferenceException("Regex.Match(html): " + CommonDefs.RegExprBody)

let GetHtmlWithNavDivCurrentroot(html: string): string = 
    if (String.IsNullOrEmpty(html)) then
        html
    else
        //let regExprDivClsNavDivCurrentroot = "<\s*div\s+\w*class=\"(nav_div_currentroot)\"(\s[^>]*)?>[\s\S]*?<\s*/div\s*>"
        let reDivs = new Regex(CommonDefs.RegExprDivClsNavDivCurrentroot, RegexOptions.IgnoreCase)
        let matchColDivs = reDivs.Matches(html)
        let lstIndexesDivs = [for m in matchColDivs -> m.Groups.Item(0).Index]
        if (lstIndexesDivs.Length > 1) then
            html.Substring(lstIndexesDivs.Item(0), lstIndexesDivs.Item(lstIndexesDivs.Length - 1) - lstIndexesDivs.Item(0))
        else if (lstIndexesDivs.Length > 0) then
            matchColDivs.Item(0).Value  
        else
            String.Empty 
            //raise <| new ArgumentOutOfRangeException("GetHtmlWithNavDivCurrentroot() : lstIndexesDivs.Length = " + lstIndexesDivs.Length.ToString()) 

let GetHtmlWithLinks(html: string): string = 
    if (String.IsNullOrEmpty(html)) then
        html
    else
        //let regExrDivs = "<\s*div\s+\w*class=\"(toclevel1\s+\w*|communityContentNavigation)\"(\s[^>]*)?>[\s\S]*?<\s*/div\s*>"
        let reDivs = new Regex(CommonDefs.RegExprDivClsToclevel1OrCommunityContentNavigation, RegexOptions.IgnoreCase)
        let matchColDivs = reDivs.Matches(html)
        let lstIndexesDivs = [for m in matchColDivs -> m.Groups.Item(0).Index]
        if (lstIndexesDivs.Length > 1) then
            html.Substring(lstIndexesDivs.Item(0), lstIndexesDivs.Item(lstIndexesDivs.Length - 1) - lstIndexesDivs.Item(0))
        else if (lstIndexesDivs.Length > 0) then
            matchColDivs.Item(0).Value 
        else
            raise <| new ArgumentOutOfRangeException("GetHtmlWithLinks() : lstIndexesDivs.Length = " + lstIndexesDivs.Length.ToString()) 

let GetHtmlTopicContainer(html: string): string = 
    if (String.IsNullOrEmpty(html)) then
        html
    else
        //let regExrDivs = "<\s*div\s+\w*class=\"(topicContainer|topicEndLine|footer)\w*\"(\s[^>]*)?>[\s\S]*?<\s*/div\s*>"
        let reDivs = new Regex(CommonDefs.RegExprDivClsTopicContainerOrTopicEndLineOrFooter, RegexOptions.IgnoreCase)
        let matchColDivs = reDivs.Matches(html)
        let lstIndexesDivs = [for m in matchColDivs -> m.Groups.Item(0).Index]
        if (lstIndexesDivs.Length >= 2) then
            let mutable txtHtml = html.Substring(lstIndexesDivs.Head, lstIndexesDivs.Tail.Head - lstIndexesDivs.Head)
            //let regExrScriptWithSrc = "<\s*script\s+\w*src=\"[\s\S]*?\"(\s[^>]*)?>[\s\S]*?<\s*/script\s*>"
            //let regExrScript = "<\s*script(\s[^>]*)?>[\s\S]*?<\s*/script\s*>"
            let reScrips = new Regex(CommonDefs.RegExprScript, RegexOptions.IgnoreCase) //RegExprScriptWithSrc
            let matchColScrips = reScrips.Matches(html)
            for m in matchColScrips do txtHtml <- txtHtml + CommonDefs.StrNewLine + m.Groups.Item(0).Value
            txtHtml
        else
            raise <| new ArgumentOutOfRangeException("GetHtmlTopicContainer(): lstIndexesDivs.Length = " + lstIndexesDivs.Length.ToString() + " (!= 2)") 

let GetListLinks(purl: string, htmlLinksPart: string): (string*string) list =
    if (String.IsNullOrEmpty(htmlLinksPart)) then
        []
    else
        // let regExrLinks = "<\s*a(\s[^>]*)?>[\s\S]*?<\s*/a\s*>"
        let reLinks = new Regex(CommonDefs.RegExprAContent, RegexOptions.IgnoreCase)
        let matchColLinks = reLinks.Matches(htmlLinksPart)
        // let regExrHref = "href\s*=\s*\"([^\"]*)\""
        // let regExrTitle = "title\s*=\s*\"([^\"]*)\""
        [for m in matchColLinks do 
            let mh = Regex.Match(m.Groups.Item(0).Value, CommonDefs.RegExprAttrHref)
            let mt = Regex.Match(m.Groups.Item(0).Value, CommonDefs.RegExprAttrTitle)
            if (mh.Success 
                && not(mh.Groups.Item("1").Captures.Item(0).Value.Contains("#")) 
                && not(String.IsNullOrEmpty(mh.Groups.Item("1").Captures.Item(0).Value)) 
                && mt.Success) then 
                    yield (RemoveBracesContentsFromUrl(CorrectUrl(purl, mh.Groups.Item("1").Captures.Item(0).Value)),
                            ApplyReplaceLstExprs(mt.Groups.Item("1").Captures.Item(0).Value, 
                                CommonDefs.RegExprLstHtmlRepl))] 

let GetListLinksEx(htmlLinksPart: string): (int*int*string*string) list =
    let mutable retVal = []
    if (not(String.IsNullOrEmpty(htmlLinksPart))) then
        //let regExrLinks = "<\s*a(\s[^>]*)?>[\s\S]*?<\s*/a\s*>"
        let reLinks = new Regex(CommonDefs.RegExprAContent, RegexOptions.IgnoreCase)
        let matchColLinks = reLinks.Matches(htmlLinksPart)
        //let regExrHref = "href\s*=\s*\"([^\"]*)\""
        for m in matchColLinks do
            try
                if (m.Success && m.Groups.Count > 1) then 
                    let matchAttrHref = Regex.Match(m.Groups.Item("1").Captures.Item(0).Value,
                                                    CommonDefs.RegExprAttrHref, 
                                                    RegexOptions.IgnoreCase)
                    if (matchAttrHref.Success && matchAttrHref.Groups.Count > 1) then
                        retVal <- (m.Groups.Item(0).Index, 
                            m.Groups.Item(0).Length, 
                            m.Groups.Item(0).Value, 
                            matchAttrHref.Groups.Item("1").Captures.Item(0).Value)::retVal
            with
                | _ -> ()
    retVal

let GetListStyles(html: string): (int*int*string*string) list =
    let mutable retVal = []
    if (not(String.IsNullOrEmpty(html))) then
        //let regExprLink = "<\s*link(\s[^>]*)?>[\s\S]*?<\s*/link\s*>"
        let reLinks = new Regex(CommonDefs.RegExprLink, RegexOptions.IgnoreCase)
        let matchColLinks = reLinks.Matches(html)
        //let regExrHref = "href\s*=\s*\"([^\"]*)\""
        for m in matchColLinks do
            try 
                if (m.Success && m.Groups.Count > 1) then 
                    let matchAttrHref = Regex.Match(m.Groups.Item("1").Captures.Item(0).Value,
                                                    CommonDefs.RegExprAttrHref, 
                                                    RegexOptions.IgnoreCase)
                    if (matchAttrHref.Success && matchAttrHref.Groups.Count > 1) then
                        retVal <- (m.Groups.Item(0).Index, 
                            m.Groups.Item(0).Length, 
                            m.Groups.Item(0).Value, 
                            matchAttrHref.Groups.Item("1").Captures.Item(0).Value)::retVal
            with
                | _ -> ()
    retVal

let GetListSrc(html: string): (int*int*string*string) list =
    let mutable retVal = []
    if (not(String.IsNullOrEmpty(html))) then
        //let regExprSrc = "src\s*=\s*\"([^\"]*)\""
        let reSrcs = new Regex(CommonDefs.RegExprSrc, RegexOptions.IgnoreCase)
        let matchSrcs = reSrcs.Matches(html)
        for m in matchSrcs do
            try 
                if (m.Success && m.Groups.Count > 1) then 
                    retVal <- (m.Groups.Item(0).Index, 
                        m.Groups.Item(0).Length, 
                        m.Groups.Item(0).Value, 
                        m.Groups.Item("1").Captures.Item(0).Value)::retVal
            with
                | _ -> ()
    retVal

let GetScripts(html: string): (bool*int*int*string*string) list =
    let mutable retVal = []
    if (not(String.IsNullOrEmpty(html))) then
        //let regExprScript = "<\s*script(\s[^>]*)?>[\s\S]*?<\s*/script\s*>"
        let reScripts = new Regex(CommonDefs.RegExprScript, RegexOptions.IgnoreCase)
        let matchColreScripts = reScripts.Matches(html)
        //let regExprSrc = "src\s*=\s*\"([^\"]*)\""
        for m in matchColreScripts do
            try 
                if (m.Success && m.Groups.Count > 1) then 
                    let matchAttrSrc = Regex.Match(m.Groups.Item("1").Captures.Item(0).Value,
                                                    CommonDefs.RegExprSrc, 
                                                    RegexOptions.IgnoreCase)
                    if (matchAttrSrc.Success && matchAttrSrc.Groups.Count > 1) then
                        retVal <- (true, 
                            m.Groups.Item(0).Index, 
                            m.Groups.Item(0).Length, 
                            m.Groups.Item(0).Value, 
                            matchAttrSrc.Groups.Item("1").Captures.Item(0).Value)::retVal
                    else
                        retVal <- (false, 
                            m.Groups.Item(0).Index, 
                            m.Groups.Item(0).Length, 
                            m.Groups.Item(0).Value, 
                            m.Groups.Item("1").Captures.Item(0).Value)::retVal
            with
                | _ -> ()
    retVal

let IsArgUrlAddress(arg: string): bool = 
    Regex.Match(arg, CommonDefs.RegExprHttp).Success 

let IsArgAspAddress(arg: string): bool = 
    Regex.Match(arg, CommonDefs.RegExprAspExt).Success

let GetHtmlAndLinksFromSource(source: string, lang: string): string*((string*string) list)*((string*string) list) = 
    let html = GetUrlContents(source, lang)
    ((html |> GetHtmlBeforeBody) + "<body>" + (html |> GetHtmlTopicContainer) + "</body></html>",
        GetListLinks(source, GetHtmlWithLinks(html)),
        GetListLinks(source, GetHtmlWithNavDivCurrentroot(html)))

let GetHtmlAndLinksFromHtml(source: string, html: string): string*((string*string) list)*((string*string) list) = 
    ((html |> GetHtmlBeforeBody) + "<body>" + (html |> GetHtmlTopicContainer) + "</body></html>",
        GetListLinks(source, GetHtmlWithLinks(html)),
        GetListLinks(source, GetHtmlWithNavDivCurrentroot(html)))

let ReplaceLinksToLocalFile(folder: string, lstUrlToReplace: (string*string*string) list (*url: string, localFile: string*), httpFile: string): unit =
    if (not <| String.IsNullOrEmpty(httpFile)) then
        let fileHttp = Path.Combine(folder, httpFile)
        use reader = new IO.StreamReader(fileHttp)
        let mutable html = reader.ReadToEnd()
        reader.Close()
        //let regExprAWithHrefContentFormat = "<\s*a\s+(\s\S[^>])*?href=\"({0})\"(\s[^>]*)?>[\s\S]*?<\s*/a\s*>"
        let len1ImgTagExternal = CommonDefs.ImgTagExternal.Length + 1
        let mutable cntChanges = 0 
        for (url, canonicalUrl, localFile) in lstUrlToReplace do
            for curUrl in [url; canonicalUrl] do
                if (not(String.IsNullOrEmpty(curUrl))) then
                    let urlArg = Regex.Replace(Regex.Replace(curUrl, "\(", "\\("), "\)", "\\)") 
                    let regExprAWithHrefContent = String.Format(CommonDefs.RegExprAWithHrefContentFormat, urlArg) 
                    let reAWithHrefContent = new Regex(regExprAWithHrefContent, RegexOptions.IgnoreCase)
                    let matchsAWithHrefContent = reAWithHrefContent.Matches(html)
                    if (matchsAWithHrefContent.Count > 0) then
                        cntChanges <- matchsAWithHrefContent.Count + cntChanges
                        for i = matchsAWithHrefContent.Count - 1 downto 0 do
                            let matchAWithHrefContent = matchsAWithHrefContent.Item(i);
                            let htmlRest = html.Substring(matchAWithHrefContent.Index + matchAWithHrefContent.Length);
                            html <- html.Substring(0, matchAWithHrefContent.Index) 
                                + Regex.Replace(matchAWithHrefContent.Value, urlArg, localFile) 
                                + Regex.Replace(htmlRest.Substring(0, len1ImgTagExternal), " " + CommonDefs.ImgTagExternal, "")
                                + htmlRest.Substring(len1ImgTagExternal)
        if (cntChanges > 0) then
            use writer = new IO.StreamWriter(Path.Combine(folder, httpFile), false, Encoding.UTF8)
            writer.Write(html)  

let ReplaceHrefsToLocalFiles(folder: string, httpFile: string, lstHtmFiles: HashSet<string>): int =
    let mutable retVal = 0
    let fileHttp = Path.Combine(folder, httpFile)
    let cfn = Path.GetFileNameWithoutExtension(httpFile).ToLower() 
    use reader = new IO.StreamReader(fileHttp)
    let mutable html = reader.ReadToEnd()
    reader.Close()
    let len1ImgTagExternal = CommonDefs.ImgTagExternal.Length + 1
    // RegExprImg = "<\s*img\s+([\s\S]*?)\s+/\s*>"
    let reImg = new Regex(CommonDefs.RegExprImg, RegexOptions.IgnoreCase)
    // RegExprAClose = "<\s*/\s*a\s*>"
    let reAClose = new Regex(CommonDefs.RegExprAClose, RegexOptions.IgnoreCase)
    // RegExprAttrHref = "href\s*=\s*\"([^\"]*)\""
    let reAttrHref = new Regex(CommonDefs.RegExprAttrHref, RegexOptions.IgnoreCase)
    // RegExprAWithHrefContent = "<\s*a\s+(\s\S[^>])*?href=\"[\s\S]*?\"(\s[^>]*)?>[\s\S]*?<\s*/a\s*>"
    let reAWithHrefContent = new Regex(CommonDefs.RegExprAWithHrefContent, RegexOptions.IgnoreCase)
    let matchsAWithHrefContents: MatchCollection = reAWithHrefContent.Matches(html)
    for i = matchsAWithHrefContents.Count - 1 downto 0 do
        let m: Match = matchsAWithHrefContents.Item(i)
        if (m.Groups.Count > 0) then
            let captureAWithHrefContent: Capture = m.Groups.Item("0").Captures.Item(0)
            let matchsAttrHrefs: MatchCollection = reAttrHref.Matches(captureAWithHrefContent.Value)
            if (matchsAttrHrefs.Count > 0 && matchsAttrHrefs.Item(0).Groups.Count > 1) then
                let captureAttrHref: Capture = matchsAttrHrefs.Item(0).Groups.Item(1).Captures.Item(0)
                if (not(captureAttrHref.Value.Contains(CommonDefs.HtmlExt)) 
                        && (Regex.IsMatch(captureAttrHref.Value, CommonDefs.RegExprHttp) 
                            || Regex.IsMatch(captureAttrHref.Value, CommonDefs.RegExprAspExt))) then
                    let fn = Path.GetFileNameWithoutExtension(Regex.Replace(captureAttrHref.Value, CommonDefs.RegExprInBraces, String.Empty)).ToLower()
                    if (fn <> cfn && not(fn.Contains("#") || fn.Contains("javascript:") || fn.Contains("?"))) then   
                        let fnHttp = Path.GetFileNameWithoutExtension(captureAttrHref.Value) + ".htm"
                        if (lstHtmFiles.Contains(fnHttp)) then  
                            html <- (html.Substring(0, captureAWithHrefContent.Index + captureAttrHref.Index)
                                + fnHttp
                                + html.Substring(captureAWithHrefContent.Index + captureAttrHref.Index + captureAttrHref.Length))
                            let posRest = captureAWithHrefContent.Index
                            let restHtml = html.Substring(posRest)
                            let matchAClose = reAClose.Match(restHtml) 
                            if (matchAClose.Success) then
                                let restHtml = restHtml.Substring(matchAClose.Index + matchAClose.Length)
                                let matchImg : Match = reImg.Match(restHtml)
                                if (matchImg.Success && matchImg.Index < len1ImgTagExternal) then
                                    if (matchImg.Value.Contains(CommonDefs.ExternalImage)) then   
                                        html <- (html.Substring(0, captureAWithHrefContent.Index + matchAClose.Index + matchAClose.Length)
                                            + restHtml.Substring(matchImg.Index + matchImg.Length))
                            retVal <- retVal + 1   
    if (retVal > 0) then
        if (File.Exists(fileHttp)) then
            File.Delete(fileHttp) 
        SaveAsFile(fileHttp, html)
    retVal

let GetDataToView(additionalLinks: XElement, folder: string, url: string, lang: string, autoReplaceUrls: bool, hsFilesInFolder: HashSet<string>, hsFilesInSubfolder: HashSet<string>): string*(string*string)list =
    let mutable lstErr = []
    if (not <| Directory.Exists(folder)) then
        raise <| new DirectoryNotFoundException(folder) 
    if (hsFilesInFolder.Count = 0) then
        Directory.GetFiles(folder, "*" + CommonDefs.HtmlExt)
        |> Array.map Path.GetFileName 
        |> hsFilesInFolder.UnionWith
        //hsFilesInFolder.UnionWith(Directory.GetFiles(folder, "*" + CommonDefs.HtmlExt)) 
    //let hsFilesInFolder = InitHashSetFilesInFolder(folder)
    let folderWithFiles = Path.Combine(folder, CommonDefs.SubfolderFiles)
    if (not <| Directory.Exists(folderWithFiles)) then
        ignore(Directory.CreateDirectory(folderWithFiles))
        if (not <| Directory.Exists(folderWithFiles)) then
            raise <| new DirectoryNotFoundException(folderWithFiles)
    let externalImageFile = Path.Combine(folderWithFiles, CommonDefs.ExternalImage)
    if (not <| File.Exists(externalImageFile)) then
        use image: System.Drawing.Image = Resoures.ModuleResoures.AppResoures.GetObject(Path.GetFileNameWithoutExtension(CommonDefs.ExternalImage)) :?> System.Drawing.Image
        image.Save(externalImageFile);
    if (not <| File.Exists(externalImageFile)) then
        raise <| new System.IO.FileNotFoundException(externalImageFile)
    if (hsFilesInSubfolder.Count = 0) then
        hsFilesInSubfolder.UnionWith(GetFilesInFolder(folderWithFiles)) 
    //let hsFilesInSubfolder = InitHashSetFilesInFolder(folderWithFiles)
    let curUrl = RemoveBracesContentsFromUrl(url)
    let mutable localFile = GetLocalFileForUrl(curUrl)
    let mutable lfHtm = Path.Combine(folder, localFile)
    if (not <| hsFilesInFolder.Contains(localFile)) then
        if (additionalLinks <> null && not(additionalLinks.IsEmpty)) then
            let canonicalHref = additionalLinks.XPathSelectElement("link/hrefCanonical[.='" + curUrl + "']")
            if (canonicalHref <> null) then
                let curUrl = canonicalHref.Parent.Element(xname "hrefNormalized").Value
                if (not(String.IsNullOrEmpty(curUrl))) then
                    localFile <- GetLocalFileForUrl(curUrl)
                    lfHtm <- Path.Combine(folder, localFile) 
    if (not <| hsFilesInFolder.Contains(localFile)) then
        let fullHtml = GetUrlContents(url, lang)
        let mutable html, lstLinks, lstNavParentLinks = GetHtmlAndLinksFromHtml(url, fullHtml)
        let normalizedUrl = GetUrlForHtmlDoc(fullHtml)
        let normalizedLocalFile = GetLocalFileForUrl(normalizedUrl)
        let lnfHtm = GetLocalFile(folder, normalizedUrl, CommonDefs.HtmlExt)
        let lstIndexesHref = GetListLinksEx(html)
        for (i,l,s,u) in lstIndexesHref do 
            let isJavaScript = u.Contains("javascript:")
            let localRef = u.Contains("#")
            let uLocalFile = if (not <| isJavaScript && not <| localRef && not <| u.Contains("?")) then 
                                GetLocalFileForUrl(u) 
                             else 
                                String.Empty
            let pLocalFile = if (not(String.IsNullOrEmpty(uLocalFile)) 
                                    && Path.GetFileNameWithoutExtension(uLocalFile).Contains(".")) then
                                let canonicalHref = additionalLinks.XPathSelectElement("link/hrefCanonical[.='" + u + "']")
                                if (canonicalHref <> null) then
                                    GetLocalFileForUrl(canonicalHref.Parent.Element(xname "hrefNormalized").Value)
                                else
                                    uLocalFile
                             else
                                uLocalFile 
            let isLocalFile = (not <| String.IsNullOrEmpty(pLocalFile) 
                                && hsFilesInFolder.Contains(pLocalFile)) 
                                || pLocalFile.Equals(localFile)
            html <- (html.Substring(0, i) 
                + if (isLocalFile) then 
                    Regex.Replace(s, u, pLocalFile) 
                  else 
                    s
                + if (not <| isLocalFile && not <| isJavaScript && not <| localRef) then 
                    " " + CommonDefs.ImgTagExternal 
                  else 
                    String.Empty   
                + html.Substring(i + l))
        let lstIndexesStyles = GetListStyles(html)
        for (i,l,s,u) in lstIndexesStyles do 
            if (not <| IsArgAspAddress(u) && u.Length > 2) then
                let regMatchRel = Regex.Match(s, CommonDefs.RegExprAttrRel)
                let attrRel =  if (regMatchRel.Success && regMatchRel.Groups.Count > 0) then regMatchRel.Groups.Item("1").Captures.Item(0).Value else String.Empty
                if (attrRel.ToLower() = "stylesheet") then 
                    let fnt =  if (u.Contains("http://")) then GetResFileName(folder, u) else Path.GetFileName(u) 
                    let fn =  if (fnt.Length > 4 && fnt.Substring(fnt.Length - 4) <> ".css") then fnt + ".css" else fnt
                    let fr = "./" + CommonDefs.SubfolderFiles + "/" + fn
                    let uq = Regex.Replace(u, CommonDefs.RegExprQChar, CommonDefs.RegExpr2QChar)
                    if (not(System.String.IsNullOrEmpty(fn)) && not <| hsFilesInSubfolder.Contains(fn)) then
                        try
                            DownloadUrlToFile(url, u, Path.Combine(folder, Path.Combine(CommonDefs.SubfolderFiles, fn)))
                            ignore(hsFilesInSubfolder.Add(fn)) 
                            html <- (html.Substring(0, i)
                                + Regex.Replace(s, uq, fr)
                                + html.Substring(i + l)) 
                        with
                            | _ as ex -> lstErr <- (u, ex.ToString())::lstErr
                    else
                        html <- (html.Substring(0, i)
                            + Regex.Replace(s, uq, fr)
                            + html.Substring(i + l))
        let lstIndexesSrc = GetListSrc(html)
        for (i,l,s,u) in lstIndexesSrc do
            if (not <| u.Contains("javascript:") (*&& not <| IsArgAspAddress(u)*) && u.Length > 2) then 
                let isJs = u.Contains("http://") && u.Contains(".js")
                let fn =  if (isJs && u.Length > 3 && u.Substring(u.Length - 3) <> ".js") then GetResFileName(folder, u) + ".js" else Path.GetFileName(u)
                let fnt = fn.ToLower()
                let mutable isFileNeedRemove: bool = false 
                if (CommonDefs.UseOfflineVersion) then
                    for fd in CommonDefs.HttpJScrptSrcToRemove do
                        if (not(isFileNeedRemove) && fd = fnt) then
                            isFileNeedRemove <- true
                if (not(isFileNeedRemove || fn.Length > 30)) then
                    let fr = "./" + CommonDefs.SubfolderFiles + "/" + fn
                    let uq = Regex.Replace(u, CommonDefs.RegExprQChar, CommonDefs.RegExpr2QChar) 
                    if (not(System.String.IsNullOrEmpty(fn)) && not <| hsFilesInSubfolder.Contains(fn)) then
                        try
                            DownloadUrlToFile(url, u, Path.Combine(folder, Path.Combine(CommonDefs.SubfolderFiles, fn)))
                            ignore(hsFilesInSubfolder.Add(fn)) 
                            html <- (html.Substring(0, i)
                                + Regex.Replace(s, uq, fr)
                                + html.Substring(i + l)) 
                        with
                            | _ as ex -> lstErr <- (u, ex.ToString())::lstErr
                    else         
                        html <- (html.Substring(0, i)
                            + Regex.Replace(s, uq, fr)
                            + html.Substring(i + l)) 
        let lstIndexesScrpt = GetScripts(html)
        for (c,i,l,s,u) in lstIndexesScrpt do
            if (c) then
                let isJs = u.Contains("http://") && u.Contains(".js")
                if (isJs) then
                    let fn =  if (isJs && u.Length > 3 && u.Substring(u.Length - 3) <> ".js") then GetResFileName(folder, u) + ".js" else Path.GetFileName(u)
                    let fnt = fn.ToLower()
                    let mutable isFileNeedRemove: bool = false 
                    if (CommonDefs.UseOfflineVersion) then
                        for fd in CommonDefs.HttpJScrptSrcToRemove do
                            if (not(isFileNeedRemove) && fd = fnt) then
                                isFileNeedRemove <- true
                    if (isFileNeedRemove || fnt.Length > 30) then
                        html <- (html.Substring(0, i)
                            + html.Substring(i + l))
                    else
                        let fr = "./" + CommonDefs.SubfolderFiles + "/" + fn
                        let uq = Regex.Replace(u, CommonDefs.RegExprQChar, CommonDefs.RegExpr2QChar) 
                        if (not(System.String.IsNullOrEmpty(fn)) && not <| hsFilesInSubfolder.Contains(fn)) then
                            try
                                DownloadUrlToFile(url, u, Path.Combine(folder, Path.Combine(CommonDefs.SubfolderFiles, fn)))
                                ignore(hsFilesInSubfolder.Add(fn)) 
                                html <- (html.Substring(0, i)
                                    + Regex.Replace(s, uq, fr)
                                    + html.Substring(i + l)) 
                            with
                                | _ as ex -> lstErr <- (u, ex.ToString())::lstErr
                        else         
                            html <- (html.Substring(0, i)
                                + Regex.Replace(s, uq, fr)
                                + html.Substring(i + l))
                else
                    if (CommonDefs.UseOfflineVersion && u.Length > 30) then
                        html <- (html.Substring(0, i)
                            + html.Substring(i + l))
            else
                if (CommonDefs.UseOfflineVersion) then
                    let isWebtrendsCom = s.Contains("webtrends.com")
                    if (isWebtrendsCom) then
                        html <- (html.Substring(0, i)
                            + CommonDefs.HttpScrptReplWebtrendsCom
                            + html.Substring(i + l)) 
        
        SaveAsFile(lnfHtm(*lfHtm*), html)
        let canonicalUrl = GetCanonicalUrlForHtmlDoc(fullHtml)
        if (canonicalUrl <> normalizedUrl) then 
            additionalLinks.Add(AdditionalLink(canonicalUrl, normalizedUrl).ToXml()) 
        ignore(hsFilesInFolder.Add(normalizedLocalFile(*localFile*)))
        if (autoReplaceUrls) then 
            let arrayOfFiles: string array = Array.zeroCreate hsFilesInFolder.Count 
            hsFilesInFolder.CopyTo(arrayOfFiles) 
            for htmlFile in arrayOfFiles do
                if (Path.GetExtension(htmlFile) = CommonDefs.HtmlExt) then 
                    ReplaceLinksToLocalFile(folder, [(url, canonicalUrl, normalizedLocalFile(*localFile*))], htmlFile)
        if (lstLinks.Length > 0) then
            let xmlParent = GrpLink("parent", lstNavParentLinks)
            let xmlCurrent = GrpLink("current", [lstLinks.Head])
            let xmlChild = GrpLink("child", lstLinks.Tail)
            let xmlLinks = XElement(xname "root", xmlParent.ToXml(), xmlCurrent.ToXml(), xmlChild.ToXml())
            let lfXml = GetLocalFile(folder, normalizedLocalFile(*localFile*), CommonDefs.XmlExt)
            if (File.Exists(lfXml)) then
                File.Delete(lfXml)
            use xw = new Xml.XmlTextWriter(lfXml, Encoding.UTF8)   
            xmlLinks.WriteTo(xw) 
        (lnfHtm(*lfHtm*), lstErr)
    else
        (lfHtm, lstErr)

let GetXElementValue(name: string, colXElements: seq<XElement>): string =
    let col = colXElements.Elements(xname name)
    let head = Seq.head col
    head.Value  

let rec AddChildNodes(statusStripValues: StatusStripValues,
                        ct: System.Threading.CancellationToken,  
                        treeView: System.Windows.Forms.TreeView,
                        node: System.Windows.Forms.TreeNode,
                        additionalLinks: XElement,                         
                        folder: string, 
                        depthTocLevel: int, 
                        limitChildItems: int, 
                        curDepthTocLevel: int,                        
                        file: string): unit =
    if (not <| ct.IsCancellationRequested) then
        try
            if (not(String.IsNullOrEmpty(folder)) || not(String.IsNullOrEmpty(file))) then
                let lfXml = GetLocalFile(folder, file, CommonDefs.XmlExt)
                if (File.Exists(lfXml)) then
                    let mutable currNode = null
                    let xmlFileLinks = XElement.Load(lfXml)
                    let xmlrCurrent = xmlFileLinks.Elements(xname "current").Elements(xname "link")       
                    let key = Regex.Replace(GetXElementValue("href", xmlrCurrent), CommonDefs.RegExprAspExt, String.Empty)
                    let title = ApplyReplaceLstExprs(GetXElementValue("title", xmlrCurrent), CommonDefs.RegExprLstHtmlRepl)
                    if (treeView.Nodes.Count = 0) then
                        currNode <- if (treeView.InvokeRequired) then 
                                        let args: obj[] = [|treeView; key; title|]
                                        treeView.Invoke(new DelegateAddNodeToTreeView(fun treeViewObj keyVal titleVal ->
                                            treeViewObj.Nodes.Add(keyVal, titleVal)), args) :?> System.Windows.Forms.TreeNode 
                                    else 
                                        treeView.Nodes.Add(key, title) 
                        currNode.Tag <- key
                    else
                        if (node <> null) then
                            currNode <- node
                        else
                            let findNodes = treeView.Nodes.Find(key, true)
                            if (findNodes <> null && findNodes.Length > 0) then
                                currNode <- findNodes.[0]
                            if (currNode = null) then
                                let curFile = Path.GetFileNameWithoutExtension(key)
                                if (curFile.Contains(".")) then
                                    let canonicalHref = additionalLinks.XPathSelectElement("link/hrefCanonical[.='" + key + "']")
                                    if (canonicalHref <> null) then
                                        let nKey = Regex.Replace(canonicalHref.Parent.Element(xname "hrefNormalized").Value, CommonDefs.RegExprAspExt, String.Empty)
                                        let findNodes = treeView.Nodes.Find(nKey, true)
                                        if (findNodes <> null && findNodes.Length > 0) then
                                            currNode <- findNodes.[0]
                                else
                                    let normalizedHref = additionalLinks.XPathSelectElement("link/hrefNormalized[.='" + key + "']")
                                    if (normalizedHref <> null) then
                                        let kKey = Regex.Replace(normalizedHref.Parent.Element(xname "hrefCanonical").Value, CommonDefs.RegExprAspExt, String.Empty)
                                        let findNodes = treeView.Nodes.Find(kKey, true)
                                        if (findNodes <> null && findNodes.Length > 0) then
                                            currNode <- findNodes.[0]
                        if (currNode = null) then
                            let xmlrParent = xmlFileLinks.Elements(xname "parent").Elements(xname "link")    
                            currNode <- null
                            if (Seq.length xmlrParent > 0) then
                                let iter = xmlrParent.GetEnumerator()
                                let rootKey = Regex.Replace((if (treeView.Nodes.Count > 0) then treeView.Nodes.[0].Tag :?> string else String.Empty), CommonDefs.RegExprAspExt, String.Empty) 
                                while iter.MoveNext() (*&& currNode = null*) do 
                                    let curLink = iter.Current
                                    let pkey = Regex.Replace(curLink.Element(xname "href").Value, CommonDefs.RegExprAspExt, String.Empty)  
                                    if (not(pkey = rootKey)) then
                                        let findNodes = treeView.Nodes.Find(pkey, true)
                                        if (findNodes <> null && findNodes.Length > 0) then
                                            currNode <- findNodes.[0]
                                if (currNode = null) then
                                    let findNodes = treeView.Nodes.Find(CommonDefs.RootItemRefAdditionally, true)
                                    if (findNodes <> null && findNodes.Length > 0) then
                                        currNode <- findNodes.[0]
                                    if (currNode = null) then
                                        let pkey = Regex.Replace(CommonDefs.RootItemRefAdditionally, CommonDefs.RegExprAspExt, String.Empty)
                                        let ptitle = ModuleResoures.GetAppResourceString("Additionally")
                                        currNode <- if (treeView.InvokeRequired) then 
                                                        let args: obj[] = [|treeView; pkey; ptitle|]
                                                        treeView.Invoke(new DelegateAddNodeToTreeView(fun treeViewObj keyVal titleVal ->
                                                            treeViewObj.Nodes.Add(keyVal, titleVal)), args) :?> System.Windows.Forms.TreeNode 
                                                    else 
                                                        treeView.Nodes.Add(pkey, ptitle) 
                                        currNode.Tag <- pkey
                                    if (currNode <> null) then
                                        let iter = xmlrParent.GetEnumerator()
                                        let mutable newNode = null
                                        let delegateAddNodeToNode = new DelegateAddNodeToNode(fun currNodeObj keyVal titleVal ->
                                            currNodeObj.Nodes.Add(keyVal, titleVal)) 
                                        while iter.MoveNext() do 
                                            let curLink = iter.Current
                                            let pkey = Regex.Replace(curLink.Element(xname "href").Value, CommonDefs.RegExprAspExt, String.Empty)  
                                            if (not(pkey = rootKey)) then
                                                let findNodes = treeView.Nodes.Find(pkey, true)
                                                if (findNodes.Length = 0) then
                                                    let ptitle = ApplyReplaceLstExprs(curLink.Element(xname "title").Value, CommonDefs.RegExprLstHtmlRepl)
                                                    newNode <- if (treeView.InvokeRequired) then 
                                                                    let args: obj[] = [|currNode; pkey; ptitle|]
                                                                    treeView.Invoke(delegateAddNodeToNode, args) :?> System.Windows.Forms.TreeNode  
                                                                  else
                                                                    currNode.Nodes.Add(pkey, ptitle) 
                                                    newNode.Tag <- pkey
                                                else
                                                    newNode <- findNodes.[0]
                                        if (newNode <> null) then
                                            currNode <- if (treeView.InvokeRequired) then 
                                                            let args: obj[] = [|newNode; key; title|]
                                                            treeView.Invoke(delegateAddNodeToNode, args) :?> System.Windows.Forms.TreeNode  
                                                        else
                                                            newNode.Nodes.Add(key, title) 
                                            currNode.Tag <- key
                            else
                                raise <| new ArgumentException(key + " " + ModuleResoures.GetAppResourceString("PhraseNoInTree"))
                
                    (*=======*)
                
                    if (currNode = null) then
                        raise <| new NullReferenceException("currNode")
                    let delegateAddNodeToNode = new DelegateAddNodeToNode(fun currNodeObj keyVal titleVal ->
                        currNodeObj.Nodes.Add(keyVal, titleVal)) 
                    let mutable isCanceled = false
                    let xmlrChild = xmlFileLinks.Elements(xname "child").Elements(xname "link")
                    let mutable curHref = null
                    for nodeLink in xmlrChild do 
                        if (nodeLink <> null) then
                            curHref <- Regex.Replace(nodeLink.Element(xname "href").Value, CommonDefs.RegExprAspExt, String.Empty)
                            let canonicalHref = if (not(additionalLinks.IsEmpty)) then additionalLinks.XPathSelectElement("link/hrefCanonical[.='" + curHref + "']") else null
                            if (canonicalHref <> null) then
                                curHref <- Regex.Replace(canonicalHref.Parent.Element(xname "hrefNormalized").Value, CommonDefs.RegExprAspExt, String.Empty)
                            if (treeView.Nodes.Find(curHref, true).Length = 0) then 
                                if ((canonicalHref = null || treeView.Nodes.Find(canonicalHref.Value, true).Length = 0)) then
                                    let title = ApplyReplaceLstExprs(nodeLink.Element(xname "title").Value, CommonDefs.RegExprLstHtmlRepl)
                                    let newNode = if (treeView.InvokeRequired) then 
                                                    let args: obj[] = [|currNode; curHref; title|]
                                                    treeView.Invoke(delegateAddNodeToNode, args) :?> System.Windows.Forms.TreeNode  
                                                  else
                                                    currNode.Nodes.Add(curHref, title) 
                                    newNode.Tag <-curHref
                                    if (statusStripValues <> null) then
                                        statusStripValues.CurCount <- statusStripValues.CurCount + 1
                                        statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusAddNodeToFile") + try " - " + Path.GetFileName(curHref) with _ -> String.Empty // "Добавлен узел для файла"
                                        if (statusStripValues.CurCount <= statusStripValues.MaxCount) then
                                            statusStripValues.ProgressBarValue <- (int)(statusStripValues.CurCount * 100) / statusStripValues.MaxCount
                                        else 
                                            if (statusStripValues.ProgressBarValue <> 0) then
                                                statusStripValues.ProgressBarValue <- 0
                                        Application.DoEvents() 
                            else 
                                if (statusStripValues <> null) then
                                    statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusSearchData") + " - " + Path.GetFileName(curHref)  // "Поиск данных..."
                                    Application.DoEvents() 
                    if (not <| isCanceled && currNode.Nodes.Count > 0 
                        && not(currNode.FullPath.Contains(ModuleResoures.GetAppResourceString("Additionally")))
                        && (depthTocLevel <= 0 || curDepthTocLevel < depthTocLevel)) then
                            for childNode in currNode.Nodes do
                                if (childNode <> null) then 
                                    AddChildNodes(
                                        statusStripValues, 
                                        ct, 
                                        treeView,
                                        childNode,
                                        additionalLinks,                                          
                                        folder, 
                                        depthTocLevel, 
                                        limitChildItems, 
                                        curDepthTocLevel + 1, 
                                        childNode.Tag :?> string)
        with
            | :? StackOverflowException as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["AddChildNodes()", ex.ToString()]
            | _ as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["AddChildNodes()", ex.ToString()]

let GetHtmlFiles(l: (string*(string*string) list) list) =
    [for x in l do
            match x with
                | (u, e) when e = [] && u <> null -> yield u
                | (u, h::t) -> ()
                | (u, _) when u = null -> ()
                | _ -> ()]

let rec DownloadAllUrl(statusStripValues: StatusStripValues,
                        ct: System.Threading.CancellationToken,  
                        additionalLinks: XElement,
                        folder: string, 
                        htmlFile: string, 
                        lang: string,
                        autoReplaceUrls: bool, 
                        hsFilesInFolder: HashSet<string>, 
                        hsFilesInSubfolder: HashSet<string>): (string*string) list =
    let mutable downloadErrs = []
    let xmlFile = GetLocalFile(folder, htmlFile, CommonDefs.XmlExt)
    if (File.Exists(xmlFile)) then
        if (statusStripValues <> null) then
            statusStripValues.CurCount <- statusStripValues.CurCount + 1
            statusStripValues.MaxCount <- hsFilesInFolder.Count 
            statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusChkNodesToFile") + " - " + Path.GetFileName(htmlFile)
            if (statusStripValues.CurCount <= statusStripValues.MaxCount) then
                statusStripValues.ProgressBarValue <- (int)(statusStripValues.CurCount * 100) / statusStripValues.MaxCount
            else 
                if (statusStripValues.ProgressBarValue <> 0) then
                    statusStripValues.ProgressBarValue <- 0
            Application.DoEvents() 
        let xmlDoc = XElement.Load(xmlFile)
        let xmlToc = xmlDoc.Elements(xname "child")
        let lstUrl =
            xmlToc.Descendants(xname "link")
                |> List.ofSeq 
                |> List.map (fun e -> e.Element(xname "href").Value)
                |> List.map (fun e -> 
                            let canonicalHref = additionalLinks.XPathSelectElement("link/hrefCanonical[.='" + e + "']")
                            if (canonicalHref <> null) then
                                canonicalHref.Parent.Element(xname "hrefNormalized").Value
                            else
                                e)
        if (lstUrl.Length > 0) then
            let lstFilesToCheck =    
                lstUrl
                |> List.map (fun u -> 
                    let lf = GetLocalFile(folder, u, CommonDefs.HtmlExt)
                    if (not <| File.Exists(lf)) then
                        statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadUrl") + " - " + Path.GetFileName(u) // "Загрузка данных с адреса"
                        Application.DoEvents() 
                        let (fn, le) = GetDataToView(additionalLinks, folder, u, lang, autoReplaceUrls, hsFilesInFolder, hsFilesInSubfolder)
                        (u, le)        
                    else
                        (null, []))
            if (lstFilesToCheck.Length > 0) then
                let lstHtmlFiles = GetHtmlFiles(lstFilesToCheck)
                if (lstHtmlFiles.Length > 0) then
                    for fn in lstHtmlFiles do
                        if (not <| ct.IsCancellationRequested) then
                            let lstErr = DownloadAllUrl(statusStripValues,
                                            ct,  
                                            additionalLinks,
                                            folder, 
                                            fn, 
                                            lang,
                                            autoReplaceUrls, 
                                            hsFilesInFolder, 
                                            hsFilesInSubfolder)
                            if (lstErr.Length > 0) then
                                downloadErrs <- List.append downloadErrs lstErr

    downloadErrs

let DownloadAllUrls(statusStripValues: StatusStripValues,
                        ct: System.Threading.CancellationToken,  
                        additionalLinks: XElement,
                        folder: string, 
                        url: string, 
                        lang: string,
                        autoReplaceUrls: bool, 
                        hsFilesInFolder: HashSet<string>, 
                        hsFilesInSubfolder: HashSet<string>): (string*string) list =
    let mutable downloadErrs = []
    try
        let arrayOfFiles: string array = Array.zeroCreate hsFilesInFolder.Count 
        hsFilesInFolder.CopyTo(arrayOfFiles) 
        for htmlFile in arrayOfFiles do
            if (not <| ct.IsCancellationRequested) then
                let lstErr = DownloadAllUrl(statusStripValues,
                                ct,  
                                additionalLinks,
                                folder, 
                                htmlFile, 
                                lang,
                                autoReplaceUrls, 
                                hsFilesInFolder, 
                                hsFilesInSubfolder)
                if (lstErr.Length > 0) then
                    downloadErrs <- List.append downloadErrs lstErr
    with
        | :? StackOverflowException as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["DownloadAllUrls()", ex.ToString()]
        | _ as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["DownloadAllUrls()", ex.ToString()]
    downloadErrs

let DownloadAll(statusStripValues: StatusStripValues, 
                cancelltionSource: System.Threading.CancellationTokenSource,
                additionalLinks: XElement,
                folder: string, 
                url: string, 
                lang: string,
                autoReplaceUrls: bool,
                hsFilesInFolder: HashSet<string>, 
                hsFilesInSubfolder: HashSet<string>) =
    let asyncOp =
        async {
                try
                    use! holder = Async.OnCancel(fun () -> statusStripValues.OperationCompleted <- true) 
                    statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadRootPage") //"Загрузка корневой страницы"
                    let (lfHtm, lstErr) = GetDataToView(additionalLinks, folder, url, lang, autoReplaceUrls, hsFilesInFolder, hsFilesInSubfolder)
                    if (lstErr.Length > 0) then    
                        statusStripValues.ListErrors <- lstErr
                    if (File.Exists(lfHtm)) then
                        statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadTOC") //"Загрузка оглавления"
                        statusStripValues.MaxCount <- hsFilesInFolder.Count //Directory.GetFiles(folder, "*" + CommonDefs.XmlExt).Length - 1
                        Application.DoEvents() 
                        let downloasErrs = DownloadAllUrls(statusStripValues, cancelltionSource.Token, additionalLinks, folder, url, lang, autoReplaceUrls, hsFilesInFolder, hsFilesInSubfolder)
                        statusStripValues.OperationCompleted <- true
                        if (downloasErrs.Length > 0) then
                            statusStripValues.ListErrors <- statusStripValues.ListErrors @ downloasErrs
                with
                | :? StackOverflowException as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["DownloadAll()", ex.ToString()]
                | _ as ex -> statusStripValues.ListErrors <- ("DownloadAll()", ex.ToString())::statusStripValues.ListErrors
                statusStripValues.OperationCompleted <- true
              }
    Async.Start(asyncOp, cancelltionSource.Token)            

let LoadTOC(statusStripValues: StatusStripValues,
            cancelltionSource: System.Threading.CancellationTokenSource,
            treeView: System.Windows.Forms.TreeView, 
            additionalLinks: XElement,
            folder: string, 
            depthTocLevel: int, 
            limitChildItems: int, 
            url: string, 
            lang: string,
            autoReplaceUrls: bool,
            hsFilesInFolder: HashSet<string>, 
            hsFilesInSubfolder: HashSet<string>): unit =
    let asyncOp =
        async {
                try
                    use! holder = Async.OnCancel(fun () -> statusStripValues.OperationCompleted <- true) 
                    statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadRootPage") //"Загрузка корневой страницы"
                    let (lfHtm, lstErr) = GetDataToView(additionalLinks, folder, url, lang, autoReplaceUrls, hsFilesInFolder, hsFilesInSubfolder)
                    if (lstErr.Length > 0) then
                        statusStripValues.ListErrors <- statusStripValues.ListErrors @ lstErr
                    if (File.Exists(lfHtm)) then
                        statusStripValues.CurCount <- 0
                        statusStripValues.MaxCount <- hsFilesInFolder.Count //Directory.GetFiles(folder, "*" + CommonDefs.XmlExt).Length - 1 
                        statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadTOC") //"Загрузка оглавления"
                        Application.DoEvents() 
                        AddChildNodes(
                            statusStripValues, 
                            cancelltionSource.Token, 
                            treeView,
                            null, 
                            additionalLinks,                             
                            folder, 
                            depthTocLevel, 
                            limitChildItems,
                            0, 
                            url)
                with
                | :? StackOverflowException as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["LoadTOC()", ex.ToString()]
                | _ as ex -> statusStripValues.ListErrors <- ("LoadTOC()", ex.ToString())::statusStripValues.ListErrors
                statusStripValues.OperationCompleted <- true
              }
    Async.Start(asyncOp, cancelltionSource.Token)            

let LoadAllChildNodes(statusStripValues: StatusStripValues,
                        cancelltionSource: System.Threading.CancellationTokenSource,
                        treeView: System.Windows.Forms.TreeView,
                        additionalLinks: XElement,
                        folder: string, 
                        file: string,
                        depthTocLevel: int, 
                        limitChildItems: int, 
                        url: string, 
                        lang: string,
                        autoReplaceUrls: bool,
                        hsFilesInFolder: HashSet<string>, 
                        hsFilesInSubfolder: HashSet<string>): unit =
    let asyncOp =
        async {
                try
                    use! holder = Async.OnCancel(fun () -> statusStripValues.OperationCompleted <- true) 
                    statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadRootPage") //"Загрузка корневой страницы"
                    let (lfHtm, lstErr) = GetDataToView(additionalLinks, folder, url, lang, autoReplaceUrls, hsFilesInFolder, hsFilesInSubfolder)
                    if (lstErr.Length > 0) then
                        statusStripValues.ListErrors <- statusStripValues.ListErrors @ lstErr
                    if (File.Exists(lfHtm)) then
                        statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadTOC") //"Загрузка оглавления"
                        statusStripValues.MaxCount <- hsFilesInFolder.Count //Directory.GetFiles(folder, "*" + CommonDefs.XmlExt).Length - 1 
                        Application.DoEvents() 
                        AddChildNodes(
                            statusStripValues, 
                            cancelltionSource.Token, 
                            treeView,
                            null, 
                            additionalLinks, 
                            folder, 
                            depthTocLevel, 
                            limitChildItems,
                            0, 
                            file)
                with
                | :? StackOverflowException as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["LoadAllChildNodes()", ex.ToString()]
                | _ as ex -> statusStripValues.ListErrors <- ("LoadAllChildNodes()", ex.ToString())::statusStripValues.ListErrors
                statusStripValues.OperationCompleted <- true
              }
    Async.Start(asyncOp, cancelltionSource.Token)

let BrowseAll(statusStripValues: StatusStripValues,
                        cancelltionSource: System.Threading.CancellationTokenSource,
                        treeView: System.Windows.Forms.TreeView,
                        additionalLinks: XElement,
                        folder: string, 
                        depthTocLevel: int, 
                        limitChildItems: int, 
                        url: string, 
                        lang: string,
                        autoReplaceUrls: bool,
                        hsFilesInFolder: HashSet<string>, 
                        hsFilesInSubfolder: HashSet<string>): unit =
    let asyncOp =
        async {
                try
                    use! holder = Async.OnCancel(fun () -> statusStripValues.OperationCompleted <- true) 
                    statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadRootPage") //"Загрузка корневой страницы"
                    let (lfHtm, lstErr) = GetDataToView(additionalLinks, folder, url, lang, autoReplaceUrls, hsFilesInFolder, hsFilesInSubfolder)
                    if (lstErr.Length > 0) then
                        statusStripValues.ListErrors <- statusStripValues.ListErrors @ lstErr
                    if (File.Exists(lfHtm)) then
                        statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusDwnLoadTOC") //"Загрузка оглавления"
                        statusStripValues.MaxCount <- hsFilesInFolder.Count //Directory.GetFiles(folder, "*" + CommonDefs.XmlExt).Length - 1 
                        Application.DoEvents() 
                        AddChildNodes(
                            statusStripValues, 
                            cancelltionSource.Token, 
                            treeView,
                            null, 
                            additionalLinks, 
                            folder, 
                            depthTocLevel, 
                            limitChildItems,
                            0, 
                            lfHtm)
                with
                | :? StackOverflowException as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["BrowseAll()", ex.ToString()]
                | _ as ex -> statusStripValues.ListErrors <- ("BrowseAll()", ex.ToString())::statusStripValues.ListErrors
                statusStripValues.OperationCompleted <- true
              }
    Async.Start(asyncOp, cancelltionSource.Token)

let ReplaceAll(statusStripValues: StatusStripValues,
                        cancelltionSource: System.Threading.CancellationTokenSource,
                        additionalLinks: XElement,
                        folder: string, 
                        url: string,
                        hsFilesInFolder: HashSet<string>): unit =
    let asyncOp =
        async {
                try
                    use! holder = Async.OnCancel(fun () -> statusStripValues.OperationCompleted <- true) 
                    let lfHtm = GetLocalFile(folder, url, CommonDefs.HtmlExt)
                    if (File.Exists(lfHtm)) then
                        statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusSearchData") + " - " + folder
                        //let htmlFiles = Directory.GetFiles(folder, "*" + CommonDefs.HtmlExt) |> Array.map (fun f -> Path.GetFileName f)
                        if (hsFilesInFolder.Count = 0) then
                            Directory.GetFiles(folder, "*" + CommonDefs.HtmlExt)
                            |> Array.map Path.GetFileName 
                            |> hsFilesInFolder.UnionWith
                        statusStripValues.MaxCount <- hsFilesInFolder.Count
                        statusStripValues.CurCount <- 0
                        let GetParentUrl(url: string): string =
                            let pos = url.LastIndexOf('/')
                            if (pos > (-1)) then
                                url.Substring(0, pos + 1)
                            else
                                "" 
                        let path = GetParentUrl(url)
                        let arrayOfFiles: string array = Array.zeroCreate hsFilesInFolder.Count 
                        hsFilesInFolder.CopyTo(arrayOfFiles) 
                        Application.DoEvents() 
                        for htmlFile in arrayOfFiles do
                            statusStripValues.CurCount <- statusStripValues.CurCount + 1
                            statusStripValues.StatusLabelValue <- ModuleResoures.GetAppResourceString("StatusReplaceUrlToHtm") + " - " + htmlFile //"Замена ссылок на локальные файлы"
                            if (statusStripValues.CurCount <= statusStripValues.MaxCount) then
                                statusStripValues.ProgressBarValue <- (int)(statusStripValues.CurCount * 100) / statusStripValues.MaxCount
                            else 
                                if (statusStripValues.ProgressBarValue <> 0) then
                                    statusStripValues.ProgressBarValue <- 0
                            Application.DoEvents() 
                            ignore(ReplaceHrefsToLocalFiles(folder, htmlFile, hsFilesInFolder))
                            (*
                            let lstUrlToReplace = [for curHtm in htmlFiles -> 
                                                    let curUrl = path + Path.GetFileNameWithoutExtension(curHtm) + ".aspx"
                                                    let normalizedHref = if (not(additionalLinks.IsEmpty)) 
                                                                            then additionalLinks.XPathSelectElement("link/hrefNormalized[.='" + curUrl + "']") 
                                                                            else null
                                                    let hrefCanonical = if (normalizedHref <> null) 
                                                                            then normalizedHref.Parent.Element(xname "hrefCanonical").Value
                                                                            else null
                                                    (curUrl, hrefCanonical, curHtm)]
                            ReplaceLinksToLocalFile(folder, lstUrlToReplace, htmlFile)
                            *)
                with
                | :? StackOverflowException as ex -> if (statusStripValues <> null) then statusStripValues.ListErrors <- ["ReplaceAll()", ex.ToString()]
                | _ as ex -> statusStripValues.ListErrors <- ("ReplaceAll()", ex.ToString())::statusStripValues.ListErrors
                statusStripValues.OperationCompleted <- true
              }
    Async.Start(asyncOp, cancelltionSource.Token)